// Pingus - A free Lemmings clone
// Copyright (C) 1999 Ingo Ruhnke <grumbel@gmx.de>
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

#include "globals.hh"
#include "Client.hh"
#include "ButtonPanel.hh"
#include "FadeOut.hh"
#include "Playfield.hh"
#include "LevelInterrupt.hh"
#include "TimeDisplay.hh"
#include "PingusCursor.hh"
#include "PinguInfo.hh"
#include "Result.hh"
#include "DebugGui.hh"
#include "PingusCounter.hh"

Client::Client(Server* s)
{
  server = s;
  fast_forward = false;
  skip_frame = 0;
}

Result
Client::start(string filename, PingusGameMode m)
{
  Result result;
  mode = m;

  do {
    result = play_level(filename);
  } while (result.next_action().get_status() == LevelInterrupt::restart);

  FadeOut::random();

  // Pingus is now going down, and maybe restart which a newly
  // selected level
  return result;
}

Result
Client::play_level(string filename)
{
  vector<GuiObj* > obj;
 
  if (verbose) {
    cout << "Constructing Window Objs: " << filename << endl;
    cout << "Starting Level: '" << filename << "'" << endl;
  }

  Result result;
  PLF*         plf          = new PLF(filename);

  server->restart(plf, result);

  //  Window*      game         = new Window();
  // Create GuiObj for Window* game 
  Playfield*    playfield    = new Playfield(plf, server->get_world());
  ButtonPanel*  button_panel = new ButtonPanel(plf);
  PinguInfo*    pingu_info   = new PinguInfo();
  DebugGui*     debug_gui    = new DebugGui();
  PingusCursor* cursor       = new PingusCursor();
  PingusCounter* pcounter    = new PingusCounter();
  TimeDisplay*  time_display = new TimeDisplay();

  debug_gui->set_world(server->get_world());
  button_panel->set_server(server);
  time_display->set_server(server);
  button_panel->set_client(this);
  pcounter->set_world(server->get_world());

  if (play_demo)
    server->set_demo(demo_file);
  
  if (record_demo)
    server->set_record_file(demo_file);
    
  if (verbose)
    cout << "Constructing Window Objs finished" << endl;

  if (music_enabled) {
    // Constructing Pingus Music
    music = new PingusMusic(plf);
    //  music->play();
  }

  playfield->set_clip_rect(0, 0, 
			   CL_Display::get_width(),
			   CL_Display::get_height() - 60);

  // Connect the button_panel with the playfield
  playfield->set_buttons(button_panel);
  playfield->set_pingu_info(pingu_info);
  playfield->set_server(server);

  if (verbose)
    cout << "Entering main_loop. Startup time: " 
	 << CL_System::get_time() << " msec." << endl;

    // Adding all GuiObj's to the screen
  obj.push_back(button_panel);
  obj.push_back(pingu_info);
  obj.push_back(playfield);
  obj.push_back(debug_gui);
  obj.push_back(pcounter);
  obj.push_back(time_display);

  if (cursor_enabled)
    obj.push_back(cursor);

  // Clear both buffers
  CL_Display::clear_display();
  CL_Display::flip_display();
  CL_Display::clear_display();
    
  // This loop will never exit, only throu a throw/catch pair
  // BUG: I don't like this part of the source, maybe somebody has an
  // idea how to rewrite it?
  try {
    int last_time = 0;
    int fps_count = 0;
      
    // Main Game Loop
    while (true) {
      CL_System::keep_alive(); // Update all input

      server->let_move();
      if (server->is_finished()) {
	throw LevelInterrupt(LevelInterrupt::select_new);
      }

      if (CL_Keyboard::get_keycode(CL_Keyboard::KEY_D)) {
	debug_gui->toggle_debug();
	while(CL_Keyboard::get_keycode(CL_Keyboard::KEY_D))
	  CL_System::keep_alive();
      }

      // Let the window move its content
      for(vector<GuiObj*>::size_type i=0; i < obj.size(); ++i) 
	obj[i]->let_move();
      
      if (!fast_forward || skip_frame > 10) {
	skip_frame = 0;
	for(vector<GuiObj*>::size_type i=0; i < obj.size(); ++i) {
	  obj[i]->draw_clipped();
	}
	CL_Display::flip_display();

	++fps_count;
	// Print the fps rate every 2secs
	if (print_fps && (last_time + 1000 < CL_System::get_time())) {
	  cout << fps_count << "fps" << endl;
	  fps_count = 0;
	  last_time = CL_System::get_time();
	}
      } else {
	++skip_frame;
      }
    }
  }
    
  catch (LevelInterrupt i) {
    server->get_world()->write_results();
    switch (i.get_status()) {
    case LevelInterrupt::restart:
    case LevelInterrupt::quit:
    case LevelInterrupt::select_new:
      result.next_action(i.get_status());
      break;
    default:
      cout << "Undefined value for LevelInterrupt" << endl;
      break;
    }
  }
  
  // Delete all alocated objects
  // FIXME: I should delete the other objects to!
  delete button_panel;
  delete    playfield;
  delete          plf;
 
  return result;
}

void
Client::toggle_fast_forward()
{
  fast_forward = !fast_forward;
  skip_frame = 0;
}

/* EOF */

