// Pingus - A free Lemmings clone
// Copyright (C) 1999 Ingo Ruhnke <grumbel@gmx.de>
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

#include "PingusResource.hh"
#include "PingusError.hh"
#include "blitter.hh"
#include "globals.hh"

#include "PingusSpotMap.hh"

MapTileSurface::MapTileSurface()
{
  empty = true;
  needs_reload = false;
}

MapTileSurface::~MapTileSurface()
{
}

bool
MapTileSurface::is_empty()
{
  return empty;
}

void
MapTileSurface::reload(void)
{
  if (needs_reload) {
    surface->reload();
    needs_reload = false;
  }
}

void
MapTileSurface::mark_dirty()
{
  needs_reload = true;
  empty = false;
}

void
MapTileSurface::check_empty()
{
  CL_SurfaceProvider* provider = surface->get_provider();
  unsigned char* buffer;
  int lenght;

  empty = true;

  provider->lock();
  lenght = provider->get_pitch() * provider->get_height(); 
  buffer = static_cast<unsigned char*>(provider->get_surface_data());
  
  // Jumping for steps because of RGBA
  for(int i=0; i < lenght; i += 4) {
    if (buffer[i] != 0) {
      /*      cout << "Tile not empty: " 
	   << int(buffer[i + 0]) << " " 
	   << int(buffer[i + 1]) << " "
	   << int(buffer[i + 2]) << " "
	   << int(buffer[i + 3]) << " " << i << endl;*/
      empty = false;
      return;
    }
  }
  //cout << "Tile empty" << endl;
  surface->get_provider()->unlock();
}


PingusSpotMap::PingusSpotMap()
{
	#ifdef WIN32
	tile_size = 32;
	#endif
}

PingusSpotMap::PingusSpotMap(PLF* plf)
{
  #ifdef WIN32
	tile_size = 32;
  #endif
  switch(plf->get_fg().type) {
  case ResDescriptor::RESOURCE:
    throw PingusError("PingusSpotMap: ResType RESOURCE is not implemented, sorry");
    break;
  case ResDescriptor::FILE:
    load(plf);
    gen_tiles();
    break;
  default:
    throw PingusError("PingusSpotMap: Unknown resource type, bailing out");
    break;
  }
}

PingusSpotMap::~PingusSpotMap(void)
{
  delete provider;
}

void
PingusSpotMap::gen_tiles(void)
{
  if (verbose) {
    cout << "PingusSpotMap_TilewWidth: " << width / tile_size << endl;
    cout << "PingusSpotMap_TilewHeight: " << height / tile_size << endl;
  }  
  
  tile.resize(width/tile_size);
  for(TileIter i=0; i < tile.size(); ++i) {
    tile[i].resize(height/tile_size);
  }
  create_maptiles(map_surface->get_provider());
}

void
PingusSpotMap::load(PLF* plf)
{
  ResDescriptor name = plf->get_fg();
  width  = plf->get_width();
  height = plf->get_height();
  load(plf->get_fg().res_name);
}

// Load the map from a *.psm (Pingu Spot Map) file and load the surfaces
void
PingusSpotMap::load(string filename)
{
  psm_parser.parse(filename);
  psm_parser.load_surfaces();
  surfaces = psm_parser.get_surfaces();

  if ((width % tile_size) != 0) {
    cerr << "Warrning: Width is not a multible of " << tile_size << endl;
    width += (tile_size - (width % tile_size));
    cerr << "Warning: Fixing height to: " << width << endl;
  }
  cout << width;
  if ((height % tile_size) != 0) {
    cerr << "Warning: Width is not a multible of " << tile_size << endl;
    height += (tile_size - (height % tile_size));
    cerr << "Warning: Fixing height to: " << height << endl;
  }
  provider = new CL_EmptySurfaceProvider(width, height, 1, -1, RGBA8888);
  for(vector<surface_data>::size_type i=0; i < surfaces.size(); ++i) 
    put_surface(provider, surfaces[i].surface->get_provider(),
		surfaces[i].x_pos, surfaces[i].y_pos);
  
  map_surface = CL_Surface::create(provider);
}

// Draws the map with a offset, needed for scrolling
void
PingusSpotMap::draw_offset(int of_x, int of_y, float s)
{
  // FIXME:Hack
  if (CL_Keyboard::get_keycode(CL_Keyboard::KEY_U))
    map_surface->reload();
  // FIXME:Hack end
  
  if (s == 1.0) {
    //map_surface->put_screen(of_x, of_y);
    for(TileIter x=0; x < tile.size(); ++x) {
      for(TileIter y=0; y < tile[x].size(); ++y) {
	if (CL_Keyboard::get_keycode(CL_Keyboard::KEY_U))
	  tile[x][y].surface->reload();
	//cout << "drawing tile: " << x << " " << y << endl;
	if (tile[x][y].is_empty()) {
	  /*
	    CL_Display::fill_rect(x * tile_size + of_x, y * tile_size + of_y, 
	    x * tile_size + tile_size - 1 + of_x, y * tile_size + tile_size -1 + of_y,
	    1.0, 0.0, 0.0, 0.5);
	  */
	} else {
	  tile[x][y].reload();
	  tile[x][y].surface->put_screen(x * tile_size + of_x, y * tile_size + of_y);
	}
      }
    }
  } else {
    map_surface->put_screen(int(of_x * s), int(of_y * s), s, s);
  }
}

// Returns the width of the map, it is read directly from the *.psm file
int
PingusSpotMap::get_width(void)
{
  return width;
}

// Returns the height of the map, it is read directly from the *.psm file
int 
PingusSpotMap::get_height(void)
{
  return height;
}

void
PingusSpotMap::remove(int x, int y)
{
  if (x > 0 && x < width
      && y > 0 && y < height) 
    {
      unsigned char* buffer;
      provider->lock();
      buffer = static_cast<unsigned char*>(provider->get_surface_data());
      buffer[y * provider->get_pitch() + (x*4)] = 0;
      provider->unlock();   
  }
}

void 
PingusSpotMap::remove(CL_SurfaceProvider* sprovider, int x, int y)
{

  if (debug_actions) {
    cout << "Bug: Debug actions is no longer supported" << endl;
    put_surface(provider, sprovider, x, y);  
  } else {
    // Get the start tile and end tile
    int start_x = x / tile_size;
    int start_y = y / tile_size;
    int end_x = (x + sprovider->get_width()) / tile_size;
    int end_y = (y + sprovider->get_height()) / tile_size;
    /*    
	  cout << "Removing something: " 
	  << start_x << " " << start_y << " - "
	  << end_x   << " " << end_y << endl;
*/
    for(int ix = start_x; ix <= end_x; ++ix) {
      for(int iy = start_y; iy <= end_y; ++iy) {
	put_alpha_surface(static_cast<CL_EmptySurfaceProvider*>(tile[ix][iy].surface->get_provider()),
			  sprovider,
			  x - (ix * tile_size), y - (iy * tile_size));
	// cout << "At: X: " << x - (ix*tile_size) << " Y: " << y - (iy*tile_size) << endl;
	tile[ix][iy].mark_dirty();
      }
    }
  }
}

void
PingusSpotMap::put(CL_SurfaceProvider* sprovider, int x, int y)
{
  //mark_tiles_dirty(x, y, sprov->get_height(), sprov->get_width());
  // Get the start tile and end tile
  int start_x = x / tile_size;
  int start_y = y / tile_size;
  int end_x = (x + sprovider->get_width()) / tile_size;
  int end_y = (y + sprovider->get_height()) / tile_size;
    
  for(int ix = start_x; ix <= end_x; ++ix) {
    for(int iy = start_y; iy <= end_y; ++iy) {
      put_surface(static_cast<CL_EmptySurfaceProvider*>(tile[ix][iy].surface->get_provider()),
		  sprovider,
		  x - (ix * tile_size), y - (iy * tile_size));
      tile[ix][iy].mark_dirty();
    }
  }
}

// Generates a ColMap (with new) and returns a pointer to it
ColMap*
PingusSpotMap::get_colmap(void)
{
  cout << "PingusSpotMap: Staring ColMap creation" << endl;
  ColMap* col;
  unsigned char* buffer;

  // Allocate the space for the colmap buffer
  // But don't delete it, since the ColMap will do that.
  buffer = new unsigned char[width * height];

  cout << "PingusSpotMap: Clearing buffer" << endl;
  for(int i=0; i < width * height; ++i) {
    buffer[i] = 0;
  }

  // Create a empty ColMap
  col = new ColMap(buffer, width, height);

  for(vector<CL_Surface*>::size_type i2=0; i2 < surfaces.size(); ++i2) {
    //cout << "Putting surface";
    col->put(surfaces[i2].surface, surfaces[i2].x_pos, surfaces[i2].y_pos);
  }
  if (verbose) cout << "Returning ColMap" << endl;
  return col;
}

CL_Surface*
PingusSpotMap::get_surface(void)
{
  return map_surface;
}

void
PingusSpotMap::create_maptiles(CL_SurfaceProvider* prov)
{
  CL_Surface* temp_surface;

  for(TileIter x=0; x < tile.size(); ++x) {
    for(TileIter y=0; y < tile[x].size(); ++y) {
      // FIXME: Memory leak?!
      temp_surface = CL_SpriteSubarrayProvider::create(prov,
						       x * tile_size,
						       y * tile_size,
						       tile_size,
						       tile_size,
						       1, 1);
      tile[x][y].surface = convert_to_emptyprovider(temp_surface);
      tile[x][y].check_empty();
    }
  }
}

void
PingusSpotMap::mark_tiles_dirty(int x, int y, int w, int h)
{
  int tile_x1 = x / tile_size;
  int tile_y1 = y / tile_size;
  int tile_x2 = (x + w) / tile_size;
  int tile_y2 = (y + h) / tile_size;

  // FIXME: Rangechecking not implemented

  for(int tile_y = tile_y1; tile_y < tile_y2; ++tile_y) {
    for(int tile_x = tile_x1; tile_x < tile_x2; ++tile_x) {
      cout << "x: " << tile_x << " y: " << tile_y << endl;
      
    }
  }
}

/* EOF */
